@@ -21,7 +21,7 @@
 from __future__ import division
 
 from ..utils import unique_rows, generate_set_partitions, UniformJointPrior, \
-                    ProductJointPrior, IndependentJointPrior, powerset
+                    ProductJointPrior, IndependentJointPrior, powerset, MaskedBounds
 from ..error_handling import GPArgumentError
 
 import scipy
@@ -45,8 +45,8 @@ class Kernel(object):
     ----------
     num_dim : positive int
         Number of dimensions of the input data. Must be consistent with the `X`
-        and `Xstar` values passed to the :py:class:`~gptools.gaussian_process.GaussianProcess` you wish
-        to use the covariance kernel with.
+        and `Xstar` values passed to the :py:class:`~gptools.gaussian_process.GaussianProcess`
+        you wish to use the covariance kernel with. Default is 1.
     num_params : Non-negative int
         Number of parameters in the model.
     initial_params : :py:class:`Array` or other Array-like, (`num_params`,), optional
@@ -113,7 +113,7 @@ class Kernel(object):
         elif len(param_names) != self.num_params:
             raise ValueError("param_names must be a list of length num_params!")
         self.param_names = param_names
-
+        
         if num_dim < 1 or not isinstance(num_dim, (int, long)):
             raise ValueError("num_dim must be an integer > 0!")
         self.num_dim = num_dim
@@ -138,7 +138,7 @@ class Kernel(object):
                 if len(fixed_params) != num_params:
                     raise ValueError("Length of fixed_params must be equal to num_params!")
         
-        # Handle default case for parameter bounds -- set them all to (0, double_max):
+        # Handle default case for parameter bounds -- set them all to (0, 1e16):
         if param_bounds is None:
             param_bounds = num_params * [(0.0, 1e16)]
         else:
@@ -178,13 +178,13 @@ class Kernel(object):
         
         Parameters
         ----------
-        Xi : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        Xj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        ni : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `i`.
-        nj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `j`.
         hyper_deriv : Non-negative int or None, optional
             The index of the hyperparameter to compute the first derivative
@@ -208,7 +208,7 @@ class Kernel(object):
     
     def set_hyperparams(self, new_params):
         """Sets the free hyperparameters to the new parameter values in new_params.
-
+        
         Parameters
         ----------
         new_params : :py:class:`Array` or other Array-like, (len(:py:attr:`self.params`),)
@@ -217,7 +217,7 @@ class Kernel(object):
         """
         new_params = scipy.asarray(new_params, dtype=float)
         
-        if new_params.shape == self.free_params.shape:
+        if len(new_params) == len(self.free_params):
             if self.enforce_bounds:
                 for idx, new_param, bound in zip(range(0, len(new_params)), new_params, self.free_param_bounds):
                     if bound[0] is not None and new_param < bound[0]:
@@ -226,7 +226,19 @@ class Kernel(object):
                         new_params[idx] = bound[1]
             self.params[~self.fixed_params] = new_params
         else:
-            raise ValueError("Length of new_params must be %s!" % (self.params.shape,))
+            raise ValueError("Length of new_params must be %s!" % (len(self.free_params),))
+    
+    @property
+    def num_free_params(self):
+        """Returns the number of free parameters.
+        """
+        return sum(~self.fixed_params)
+    
+    @property
+    def free_param_idxs(self):
+        """Returns the indices of the free parameters in the main arrays of parameters, etc.
+        """
+        return scipy.arange(0, self.num_params)[~self.fixed_params]
     
     @property
     def free_params(self):
@@ -237,7 +249,11 @@ class Kernel(object):
         free_params : :py:class:`Array`
             Array of the free parameters, in order.
         """
-        return self.params[~self.fixed_params]
+        return MaskedBounds(self.params, self.free_param_idxs)
+    
+    @free_params.setter
+    def free_params(self, value):
+        self.params[self.free_param_idxs] = scipy.asarray(value, dtype=float)
     
     @property
     def free_param_bounds(self):
@@ -248,7 +264,13 @@ class Kernel(object):
         free_param_bounds : :py:class:`Array`
             Array of the bounds of the free parameters, in order.
         """
-        return scipy.asarray(self.param_bounds, dtype=float)[~self.fixed_params]
+        return MaskedBounds(self.hyperprior.bounds, self.free_param_idxs)
+    
+    @free_param_bounds.setter
+    def free_param_bounds(self, value):
+        # Need to use a loop since self.hyperprior.bounds is NOT guaranteed to support fancy indexing.
+        for i, v in zip(self.free_param_idxs, value):
+            self.hyperprior.bounds[i] = v
     
     @property
     def free_param_names(self):
@@ -259,7 +281,13 @@ class Kernel(object):
         free_param_names : :py:class:`Array`
             Array of the names of the free parameters, in order.
         """
-        return scipy.asarray(self.param_names)[~self.fixed_params]
+        return MaskedBounds(self.param_names, self.free_param_idxs)
+    
+    @free_param_names.setter
+    def free_param_names(self, value):
+        # Cast to array in case it hasn't been done already:
+        self.param_names = scipy.asarray(self.param_names, dtype=str)
+        self.param_names[~self.fixed_params] = value
     
     def __add__(self, other):
         """Add two Kernels together.
@@ -304,8 +332,8 @@ class Kernel(object):
         
         Parameters
         ----------
-        tau : :py:class:`Array`, (`M`, `N`)
-            `M` inputs with dimension `N`.
+        tau : :py:class:`Array`, (`M`, `D`)
+            `M` inputs with dimension `D`.
         return_l : bool, optional
             Set to True to return a tuple of (`tau`, `l_mat`), where `l_mat`
             is the matrix of length scales to match the shape of `tau`. Default
@@ -315,8 +343,8 @@ class Kernel(object):
         -------
         r2l2 : :py:class:`Array`, (`M`,)
             Anisotropically scaled distances squared.
-        l_mat : :py:class:`Array`, (`M`, `N`)
-            The (`N`,) array of length scales repeated for each of the `M`
+        l_mat : :py:class:`Array`, (`M`, `D`)
+            The (`D`,) array of length scales repeated for each of the `M`
             inputs. Only returned if `return_l` is True.
         """
         l_mat = scipy.tile(self.params[-self.num_dim:], (tau.shape[0], 1))
@@ -428,12 +456,12 @@ class BinaryKernel(Kernel):
         """
         new_params = scipy.asarray(new_params, dtype=float)
         
-        if new_params.shape == self.free_params.shape:
+        if len(new_params) == len(self.free_params):
             num_free_k1 = sum(~self.k1.fixed_params)
             self.k1.set_hyperparams(new_params[:num_free_k1])
             self.k2.set_hyperparams(new_params[num_free_k1:])
         else:
-            raise ValueError("Length of new_params must be %s!" % (self.free_params.shape,))
+            raise ValueError("Length of new_params must be %s!" % (len(self.free_params),))
 
 class SumKernel(BinaryKernel):
     """The sum of two kernels.
@@ -443,13 +471,13 @@ class SumKernel(BinaryKernel):
         
         Parameters
         ----------
-        Xi : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        Xj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        ni : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `i`.
-        nj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `j`.
         symmetric : bool, optional
             Whether or not the input `Xi`, `Xj` are from a symmetric matrix.
@@ -478,13 +506,13 @@ class ProductKernel(BinaryKernel):
         
         Parameters
         ----------
-        Xi : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        Xj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        ni : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `i`.
-        nj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `j`.
         symmetric : bool, optional
             Whether or not the input `Xi`, `Xj` are from a symmetric matrix.
@@ -546,13 +574,13 @@ class ChainRuleKernel(Kernel):
         
         Parameters
         ----------
-        Xi : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        Xj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        ni : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `i`.
-        nj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `j`.
         hyper_deriv : Non-negative int or None, optional
             The index of the hyperparameter to compute the first derivative
@@ -602,9 +630,9 @@ class ChainRuleKernel(Kernel):
 
         Parameters
         ----------
-        tau : :py:class:`Matrix`, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        n : :py:class:`Array`, (`N`,)
+        tau : :py:class:`Matrix`, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        n : :py:class:`Array`, (`D`,)
             Degree of derivative with respect to each dimension.
 
         Returns
@@ -640,8 +668,8 @@ class ChainRuleKernel(Kernel):
 
         Parameters
         ----------
-        tau : :py:class:`Matrix`, (`M`, `N`)
-            `M` inputs with dimension `N`.
+        tau : :py:class:`Matrix`, (`M`, `D`)
+            `M` inputs with dimension `D`.
         p : list of :py:class:`Array`
             Each element is a block of the partition representing the
             derivative orders to use.
@@ -722,13 +750,13 @@ class ArbitraryKernel(Kernel):
         
         Parameters
         ----------
-        Xi : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        Xj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        ni : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `i`.
-        nj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `j`.
         hyper_deriv : Non-negative int or None, optional
             The index of the hyperparameter to compute the first derivative
@@ -901,19 +929,19 @@ class MaskedKernel(Kernel):
     def __call__(self, Xi, Xj, ni, nj, **kwargs):
         """Evaluate the covariance between points `Xi` and `Xj` with derivative order `ni`, `nj`.
         
-        Note that in the argument specifications, `N` is the `total_dim`
+        Note that in the argument specifications, `D` is the `total_dim`
         specified in the constructor (i.e., :py:attr:`num_dim` for the
         :py:class:`MaskedKernel` instance itself).
         
         Parameters
         ----------
-        Xi : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        Xj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        ni : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `i`.
-        nj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `j`.
         hyper_deriv : Non-negative int or None, optional
             The index of the hyperparameter to compute the first derivative
