@@ -240,18 +240,26 @@ class GibbsKernel1d(Kernel):
         Function that dictates the length scale warping and its derivative.
         Must have fingerprint (`x`, `n`, `p1`, `p2`, ...) where `p1` is element
         one of the kernel's parameters (i.e., element zero is skipped).
+    num_params : int, optional
+        The number of parameters of the length scale function. If not passed,
+        introspection will be used to determine this. This will fail if you have
+        used the \*args syntax in your function definition. This count should
+        include sigma_f, even though it is not passed to the length scale
+        function.
     **kwargs
         All remaining arguments are passed to :py:class:`~gptools.kernel.core.Kernel`.
     """
-    def __init__(self, l_func, **kwargs):
+    def __init__(self, l_func, num_params=None, **kwargs):
         self.l_func = l_func
-        # There are two unimportant parameters at the start of the l_func
-        # fingerprint, then we have to add one for sigma_f.
-        try:
-            num_params = len(inspect.getargspec(l_func)[0]) - 2 + 1
-        except TypeError:
-            # Need to remove self from the arg list for bound method:
-            num_params = len(inspect.getargspec(l_func.__call__)[0]) - 3 + 1
+        # TODO: Some of the logic from warping should be ported over.
+        if num_params is None:
+            # There are two unimportant parameters at the start of the l_func
+            # fingerprint, then we have to add one for sigma_f.
+            try:
+                num_params = len(inspect.getargspec(l_func)[0]) - 2 + 1
+            except TypeError:
+                # Need to remove self from the arg list for bound method:
+                num_params = len(inspect.getargspec(l_func.__call__)[0]) - 3 + 1
         
         super(GibbsKernel1d, self).__init__(num_dim=1, num_params=num_params, **kwargs)
     
@@ -260,13 +268,13 @@ class GibbsKernel1d(Kernel):
         
         Parameters
         ----------
-        Xi : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        Xj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
-            `M` inputs with dimension `N`.
-        ni : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `i`.
-        nj : :py:class:`Matrix` or other Array-like, (`M`, `N`)
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
             `M` derivative orders for set `j`.
         hyper_deriv : Non-negative int or None, optional
             The index of the hyperparameter to compute the first derivative
@@ -285,10 +293,10 @@ class GibbsKernel1d(Kernel):
         ------
         NotImplementedError
             If the `hyper_deriv` keyword is not None.
-        
+        """
         if hyper_deriv is not None:
             raise NotImplementedError("Hyperparameter derivatives have not been implemented!")
-        """
+        
         n_combined = scipy.asarray(scipy.hstack((ni, nj)), dtype=int)
         n_combined_unique = unique_rows(n_combined)
         
@@ -511,8 +519,8 @@ def double_tanh_warp(x, n, lcore, lmid, ledge, la, lb, xa, xb):
     if n == 0:
         return a * scipy.tanh((x - xa) / la) + b * scipy.tanh((x - xb) / lb) + c
     elif n == 1:
-        return (a**2 / la * (scipy.cosh((x - xa) / la))**(-2.0) +
-                b**2 / lb * (scipy.cosh((x - xb) / lb))**(-2.0))
+        return (a / la * (scipy.cosh((x - xa) / la))**(-2.0) +
+                b / lb * (scipy.cosh((x - xb) / lb))**(-2.0))
     else:
         raise NotImplementedError("Only derivatives up to order 1 are supported!")
 
