@@ -20,12 +20,21 @@
 
 from __future__ import division
 
-from .core import ChainRuleKernel, ArbitraryKernel
-from ..utils import generate_set_partitions
+import warnings
+
+from .core import ChainRuleKernel, ArbitraryKernel, Kernel
+from ..utils import generate_set_partitions, unique_rows
+try:
+    from ._matern import _matern52
+except ImportError:
+    warnings.warn(
+        "Could not import _matern, the extension might not have been built "
+        "properly. The optimized Matern52Kernel will not be available.",
+        ImportWarning
+    )
 
 import scipy
 import scipy.special
-import warnings
 try:
     import mpmath
 except ImportError:
@@ -133,10 +142,117 @@ class MaternKernelArb(ArbitraryKernel):
         """
         return self.params[1]
 
+class MaternKernel1d(Kernel):
+    r"""Matern covariance kernel. Only for univariate data. Only supports up to first derivatives. Treats order as a hyperparameter.
+    
+    The Matern kernel has the following hyperparameters, always referenced in
+    the order listed:
+    
+    = ===== ===============
+    0 sigma prefactor
+    1 nu    order of kernel
+    2 l1    length scale
+    = ===== ===============
+    
+    The kernel is defined as:
+    
+    .. math::
+    
+        k_M = \sigma^2 \frac{2^{1-\nu}}{\Gamma(\nu)}
+        \left (\sqrt{2\nu \left (\frac{\tau^2}{l_1^2}\right )}\right )^\nu
+        K_\nu\left(\sqrt{2\nu \left(\frac{\tau^2}{l_1^2}\right)}\right)
+    
+    where :math:`\tau=X_i-X_j`.
+    
+    Note that the expressions for the derivatives break down for :math:`\nu < 1`.
+
+    Parameters
+    ----------
+    **kwargs
+        All keyword parameters are passed to :py:class:`~gptools.kernel.core.Kernel`.
+    """
+    def __init__(self, **kwargs):
+        param_names = [r'\sigma_f', r'\nu', r'l_1']
+        super(MaternKernel1d, self).__init__(
+            num_dim=1,
+            num_params=3,
+            param_names=param_names,
+            **kwargs
+        )
+    
+    def __call__(self, Xi, Xj, ni, nj, hyper_deriv=None, symmetric=False):
+        """Evaluate the covariance between points `Xi` and `Xj` with derivative order `ni`, `nj`.
+        
+        Parameters
+        ----------
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` derivative orders for set `i`.
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` derivative orders for set `j`.
+        hyper_deriv : Non-negative int or None, optional
+            The index of the hyperparameter to compute the first derivative
+            with respect to. If None, no derivatives are taken. Hyperparameter
+            derivatives are not supported at this point. Default is None.
+        symmetric : bool, optional
+            Whether or not the input `Xi`, `Xj` are from a symmetric matrix.
+            Default is False.
+        
+        Returns
+        -------
+        Kij : :py:class:`Array`, (`M`,)
+            Covariances for each of the `M` `Xi`, `Xj` pairs.
+        
+        Raises
+        ------
+        NotImplementedError
+            If the `hyper_deriv` keyword is not None.
+        """
+        if hyper_deriv is not None:
+            raise NotImplementedError("Hyperparameter derivatives have not been implemented!")
+        
+        n_combined = scipy.asarray(scipy.hstack((ni, nj)), dtype=int)
+        n_combined_unique = unique_rows(n_combined)
+        
+        x_y = scipy.asarray(Xi, dtype=float).ravel() - scipy.asarray(Xj, dtype=float).ravel()
+        
+        zero_mask = x_y == 0
+        
+        q = scipy.sqrt(2 * self.params[1] * x_y**2) / self.params[2]
+        
+        k = scipy.zeros(Xi.shape[0], dtype=float)
+        for n_combined_state in n_combined_unique:
+            idxs = (n_combined == n_combined_state).all(axis=1)
+            # Derviative expressions evaluated with Mathematica, assuming l>0.
+            if (n_combined_state == scipy.asarray([0, 0])).all():
+                k[idxs] = q[idxs]**self.params[1] * scipy.special.kv(self.params[1], q[idxs])
+                k[idxs & zero_mask] = 2**(self.params[1] - 1) * scipy.special.gamma(self.params[1])
+            elif (n_combined_state == scipy.asarray([1, 0])).all():
+                k[idxs] = -1.0 / x_y[idxs] * q[idxs]**(1 + self.params[1]) * scipy.special.kv(self.params[1] - 1, q[idxs])
+                k[idxs & zero_mask] = 0.0
+            elif (n_combined_state == scipy.asarray([0, 1])).all():
+                k[idxs] = 1.0 / x_y[idxs] * q[idxs]**(1 + self.params[1]) * scipy.special.kv(self.params[1] - 1, q[idxs])
+                k[idxs & zero_mask] = 0.0
+            elif (n_combined_state == scipy.asarray([1, 1])).all():
+                k[idxs] = 2.0 * self.params[1] / (self.params[2])**2 * q[idxs]**self.params[1] * (
+                    -scipy.special.kv(self.params[1] - 2, q[idxs]) +
+                    scipy.special.kv(self.params[1] - 1, q[idxs]) / q[idxs]
+                )
+                # Had to assume nu > 1 for this to work: CHECK THIS!
+                k[idxs & zero_mask] = 2**(self.params[1] - 1) * self.params[1] * scipy.special.gamma(self.params[1] - 1) / (self.params[2]**2)
+            else:
+                raise NotImplementedError("Derivatives greater than [1, 1] are not supported!")
+        k = (self.params[0]**2 * 2**(1 - self.params[1]) / scipy.special.gamma(self.params[1])) * k
+        return k
 
 class MaternKernel(ChainRuleKernel):
     r"""Matern covariance kernel. Supports arbitrary derivatives. Treats order as a hyperparameter.
     
+    EXPERIMENTAL IMPLEMENTATION! DO NOT TRUST FOR HIGHER DERIVATIVES!
+    
     The Matern kernel has the following hyperparameters, always referenced in
     the order listed:
     
@@ -184,8 +300,8 @@ class MaternKernel(ChainRuleKernel):
         
         Parameters
         ----------
-        tau : :py:class:`Matrix`, (`M`, `N`)
-            `M` inputs with dimension `N`.
+        tau : :py:class:`Matrix`, (`M`, `D`)
+            `M` inputs with dimension `D`.
         
         Returns
         -------
@@ -202,8 +318,8 @@ class MaternKernel(ChainRuleKernel):
         
         Parameters
         ----------
-        tau : :py:class:`Matrix`, (`M`, `N`)
-            `M` inputs with dimension `N`.
+        tau : :py:class:`Matrix`, (`M`, `D`)
+            `M` inputs with dimension `D`.
         return_r2l2 : bool, optional
             Set to True to return a tuple of (`y`, `r2l2`). Default is False
             (only return `y`).
@@ -317,8 +433,8 @@ class MaternKernel(ChainRuleKernel):
         
         Parameters
         ----------
-        tau : :py:class:`Matrix`, (`M`, `N`)
-            `M` inputs with dimension `N`.
+        tau : :py:class:`Matrix`, (`M`, `D`)
+            `M` inputs with dimension `D`.
         b : :py:class:`Array`, (`P`,)
             Block specifying derivatives to be evaluated.
         r2l2 : :py:class:`Array`, (`M`,)
@@ -356,8 +472,8 @@ class MaternKernel(ChainRuleKernel):
         
         Parameters
         ----------
-        tau : :py:class:`Matrix`, (`M`, `N`)
-            `M` inputs with dimension `N`.
+        tau : :py:class:`Matrix`, (`M`, `D`)
+            `M` inputs with dimension `D`.
         p : list of :py:class:`Array`
             Each element is a block of the partition representing the derivative
             orders to use.    
@@ -385,8 +501,8 @@ class MaternKernel(ChainRuleKernel):
         
         Parameters
         ----------
-            tau : :py:class:`Matrix`, (`M`, `N`)
-                `M` inputs with dimension `N`.
+            tau : :py:class:`Matrix`, (`M`, `D`)
+                `M` inputs with dimension `D`.
             b : :py:class:`Array`, (`P`,)
                 Block specifying derivatives to be evaluated.
         
@@ -413,3 +529,93 @@ class MaternKernel(ChainRuleKernel):
         r"""Returns the value of the order :math:`\nu`.
         """
         return self.params[1]
+
+
+class Matern52Kernel(Kernel):
+    r"""Matern 5/2 covariance kernel. Supports only 0th and 1st derivatives
+    and is fixed at nu=5/2. Because of these limitations, it is quite a bit
+    faster than the more general Matern kernels.
+
+    The Matern52 kernel has the following hyperparameters, always referenced in
+    the order listed:
+
+    = ===== ====================================
+    0 sigma prefactor
+    1 l1    length scale for the first dimension
+    2 l2    ...and so on for all dimensions
+    = ===== ====================================
+
+    The kernel is defined as:
+
+    .. math::
+
+        k_M(x, x') = \sigma^2 \left(1 + \sqrt{5r^2} + \frac{5}{3}r^2\right) \exp(-\sqrt{5r^2}) \\
+        r^2 = \sum_{d=1}^D \frac{(x_d - x'_d)^2}{l_d^2}
+
+    Parameters
+    ----------
+    num_dim : int
+        Number of dimensions of the input data. Must be consistent with the `X`
+        and `Xstar` values passed to the :py:class:`~gptools.gaussian_process.GaussianProcess`
+        you wish to use the covariance kernel with.
+    **kwargs
+        All keyword parameters are passed to :py:class:`~gptools.kernel.core.Kernel`.
+
+    Raises
+    ------
+    ValueError
+        If `num_dim` is not a positive integer or the lengths of the input
+        vectors are inconsistent.
+    GPArgumentError
+        If `fixed_params` is passed but `initial_params` is not.
+    """
+    def __init__(self, num_dim=1, **kwargs):
+        param_names = [r'\sigma_f'] + ['l_%d' % (i + 1,) for i in range(0, num_dim)]
+        super(Matern52Kernel, self).__init__(num_dim=num_dim,
+                                             num_params=num_dim + 1,
+                                             param_names=param_names,
+                                             **kwargs)
+    def __call__(self, Xi, Xj, ni, nj, hyper_deriv=None, symmetric=False):
+        """Evaluate the covariance between points `Xi` and `Xj` with derivative order `ni`, `nj`.
+
+        Parameters
+        ----------
+        Xi : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        Xj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` inputs with dimension `D`.
+        ni : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` derivative orders for set `i`.
+        nj : :py:class:`Matrix` or other Array-like, (`M`, `D`)
+            `M` derivative orders for set `j`.
+        hyper_deriv : Non-negative int or None, optional
+            The index of the hyperparameter to compute the first derivative
+            with respect to. If None, no derivatives are taken. Hyperparameter
+            derivatives are not supported at this point. Default is None.
+        symmetric : bool
+            Whether or not the input `Xi`, `Xj` are from a symmetric matrix.
+            Default is False.
+
+        Returns
+        -------
+        Kij : :py:class:`Array`, (`M`,)
+            Covariances for each of the `M` `Xi`, `Xj` pairs.
+
+        Raises
+        ------
+        NotImplementedError
+            If the `hyper_deriv` keyword is not None.
+        """
+        if hyper_deriv is not None:
+            raise NotImplementedError("Hyperparameter derivatives have not been implemented!")
+        if scipy.any(scipy.sum(ni, axis=1) > 1) or scipy.any(scipy.sum(nj, axis=1) > 1):
+            raise ValueError("Matern52Kernel only supports 0th and 1st order derivatives")
+
+        Xi = scipy.asarray(Xi, dtype=scipy.float64)
+        Xj = scipy.asarray(Xj, dtype=scipy.float64)
+        ni = scipy.array(ni, dtype=scipy.int32)
+        nj = scipy.array(nj, dtype=scipy.int32)
+        var = scipy.square(self.params[-self.num_dim:])
+
+        value = _matern52(Xi, Xj, ni, nj, var)
+        return self.params[0]**2 * value
