@@ -234,6 +234,25 @@ class CombinedBounds(object):
     
     def __len__(self):
         return len(self.l1) + len(self.l2)
+    
+    def __invert__(self):
+        return ~scipy.asarray(self)
+
+class MaskedBounds(object):
+    """Object to support reassignment of free parameter bounds.
+    """
+    def __init__(self, a, m):
+        self.a = a
+        self.m = m
+    
+    def __getitem__(self, pos):
+        return self.a[self.m[pos]]
+    
+    def __setitem__(self, pos, value):
+        self.a[self.m[pos]] = value
+    
+    def __len__(self):
+        return len(self.m)
 
 class ProductJointPrior(JointPrior):
     """Product of two independent priors.
@@ -399,9 +418,9 @@ class CoreMidEdgeJointPrior(UniformJointPrior):
         ll = 0
         bounds_new = copy.copy(self.bounds)
         # lc < lm:
-        bounds_new[1] = (self.bounds[1][0], theta[2])
+        # bounds_new[1] = (self.bounds[1][0], theta[2])
         # le < lm:
-        bounds_new[3] = (self.bounds[3][0], theta[2])
+        # bounds_new[3] = (self.bounds[3][0], theta[2])
         # xa < xb:
         bounds_new[6] = (self.bounds[6][0], theta[7])
         for v, b in zip(theta, bounds_new):
@@ -435,15 +454,15 @@ class CoreMidEdgeJointPrior(UniformJointPrior):
         
         out = scipy.zeros(out_shape)
         # sigma_f, lm, la, lb, xb:
-        for j in [0, 2, 4, 5, 7]:
+        for j in [0, 1, 2, 3, 4, 5, 7]:
             out[j, :] = numpy.random.uniform(low=self.bounds[j][0],
                                              high=self.bounds[j][1],
                                              size=size)
         # lc, le:
-        for j in [1, 3]:
-            out[j, :] = numpy.random.uniform(low=self.bounds[j][0],
-                                             high=out[2, :],
-                                             size=size)
+        # for j in [1, 3]:
+        #     out[j, :] = numpy.random.uniform(low=self.bounds[j][0],
+        #                                      high=out[2, :],
+        #                                      size=size)
         # xa:
         out[6, :] = numpy.random.uniform(low=self.bounds[6][0],
                                          high=out[7, :],
@@ -500,6 +519,106 @@ class IndependentJointPrior(JointPrior):
         """
         return [p.rvs(size=size) for p in self.univariate_priors]
 
+class LogNormalJointPrior(JointPrior):
+    """Joint prior for which each hyperparameter has a log-normal prior with fixed hyper-hyperparameters.
+    
+    Parameters
+    ----------
+    mu : list of float, same size as `sigma`
+        Means of the logarithms of the hyperparameters.
+    sigma : list of float
+        Standard deviations of the logarithms of the hyperparameters.
+    """
+    def __init__(self, mu, sigma):
+        sigma = scipy.atleast_1d(scipy.asarray(sigma, dtype=float))
+        mu = scipy.atleast_1d(scipy.asarray(mu, dtype=float))
+        if sigma.shape != mu.shape:
+            raise ValueError("sigma and mu must have the same shape!")
+        if sigma.ndim != 1:
+            raise ValueError("sigma and mu must both be one dimensional!")
+        self.sigma = sigma
+        self.emu = scipy.exp(mu)
+    
+    def __call__(self, theta):
+        """Evaluate the prior log-PDF at the given values of the hyperparameters, theta.
+        
+        Parameters
+        ----------
+        theta : array-like, (`num_params`,)
+            The hyperparameters to evaluate the log-PDF at.
+        """
+        ll = 0
+        for v, s, em in zip(theta, self.sigma, self.emu):
+            ll += scipy.stats.lognorm.logpdf(v, s, loc=0, scale=em)
+        return ll
+    
+    @property
+    def bounds(self):
+        """The bounds of the random variable.
+        """
+        return [scipy.stats.lognorm.interval(1, s, loc=0, scale=em) for s, em in zip(self.sigma, self.emu)]
+    
+    def random_draw(self, size=None):
+        """Draw random samples of the hyperparameters.
+        
+        Parameters
+        ----------
+        size : None, int or array-like, optional
+            The number/shape of samples to draw. If None, only one sample is
+            returned. Default is None.
+        """
+        return [scipy.stats.lognorm.rvs(s, loc=0, scale=em, size=size) for s, em in zip(self.sigma, self.emu)]
+
+class GammaJointPrior(JointPrior):
+    """Joint prior for which each hyperparameter has a gamma prior with fixed hyper-hyperparameters.
+    
+    Parameters
+    ----------
+    a : list of float, same size as `b`
+        Shape parameters.
+    b : list of float
+        Rate parameters.
+    """
+    def __init__(self, a, b):
+        a = scipy.atleast_1d(scipy.asarray(a, dtype=float))
+        b = scipy.atleast_1d(scipy.asarray(b, dtype=float))
+        if a.shape != b.shape:
+            raise ValueError("sigma and mu must have the same shape!")
+        if a.ndim != 1:
+            raise ValueError("sigma and mu must both be one dimensional!")
+        self.a = a
+        self.b = b
+    
+    def __call__(self, theta):
+        """Evaluate the prior log-PDF at the given values of the hyperparameters, theta.
+        
+        Parameters
+        ----------
+        theta : array-like, (`num_params`,)
+            The hyperparameters to evaluate the log-PDF at.
+        """
+        ll = 0
+        for v, a, b in zip(theta, self.a, self.b):
+            ll += scipy.stats.gamma.logpdf(v, a, loc=0, scale=1.0 / b)
+        return ll
+    
+    @property
+    def bounds(self):
+        """The bounds of the random variable.
+        """
+        return [scipy.stats.gamma.interval(1, a, loc=0, scale=1.0 / b) for a, b in zip(self.a, self.b)]
+    
+    def random_draw(self, size=None):
+        """Draw random samples of the hyperparameters.
+        
+        Parameters
+        ----------
+        size : None, int or array-like, optional
+            The number/shape of samples to draw. If None, only one sample is
+            returned. Default is None.
+        """
+        return [scipy.stats.gamma.rvs(a, loc=0, scale=1.0 / b, size=size) for a, b in zip(self.a, self.b)]
+
 def wrap_fmin_slsqp(fun, guess, opt_kwargs={}):
     """Wrapper for :py:func:`fmin_slsqp` to allow it to be called with :py:func:`minimize`-like syntax.
 
@@ -746,14 +865,23 @@ def powerset(iterable):
     s = list(iterable)
     return itertools.chain.from_iterable(itertools.combinations(s, r) for r in range(len(s) + 1))
 
-def unique_rows(arr):
+def unique_rows(arr, return_index=False, return_inverse=False):
     """Returns a copy of arr with duplicate rows removed.
     
     From Stackoverflow "Find unique rows in numpy.array."
     
     Parameters
     ----------
-    arr : :py:class:`Array`, (`m`, `n`). The array to find the unique rows of.
+    arr : :py:class:`Array`, (`m`, `n`)
+        The array to find the unique rows of.
+    return_index : bool, optional
+        If True, the indices of the unique rows in the array will also be
+        returned. I.e., unique = arr[idx]. Default is False (don't return
+        indices).
+    return_inverse: bool, optional
+        If True, the indices in the unique array to reconstruct the original
+        array will also be returned. I.e., arr = unique[inv]. Default is False
+        (don't return inverse).
     
     Returns
     -------
@@ -764,8 +892,17 @@ def unique_rows(arr):
         scipy.dtype((scipy.void, arr.dtype.itemsize * arr.shape[1]))
     )
     try:
-        dum, idx = scipy.unique(b, return_index=True)
+        out = scipy.unique(b, return_index=True, return_inverse=return_inverse)
+        dum = out[0]
+        idx = out[1]
+        if return_inverse:
+            inv = out[2]
     except TypeError:
+        if return_inverse:
+            raise RuntimeError(
+                "Error in scipy.unique on older versions of numpy prevents "
+                "return_inverse from working!"
+            )
         # Handle bug in numpy 1.6.2:
         rows = [_Row(row) for row in b]
         srt_idx = sorted(range(len(rows)), key=rows.__getitem__)
@@ -776,7 +913,14 @@ def unique_rows(arr):
         row_cmp = scipy.asarray(row_cmp)
         transition_idxs = scipy.where(row_cmp != 0)[0]
         idx = scipy.asarray(srt_idx)[transition_idxs]
-    return arr[idx]
+    out = arr[idx]
+    if return_index:
+        out = (out, idx)
+    elif return_inverse:
+        out = (out, inv)
+    elif return_index and return_inverse:
+        out = (out, idx, inv)
+    return out
 
 class _Row(object):
     """Helper class to compare rows of a matrix.
@@ -827,7 +971,7 @@ def compute_stats(vals, check_nan=False, robust=False, axis=1, plot_QQ=False, bi
     
     Parameters
     ----------
-    vals : array-like, (`M`, `N`)
+    vals : array-like, (`M`, `D`)
         Values to compute the average statistics along the specified axis of.
     check_nan : bool, optional
         Whether or not to check for (and exclude) NaN's. Default is False (do
@@ -969,6 +1113,121 @@ def univariate_envelope_plot(x, mean, std, ax=None, base_alpha=0.375, envelopes=
     color = plt.getp(l[0], 'color')
     e = []
     for i in envelopes:
-        e.append(ax.fill_between(x, mean - i * std, mean + i * std,
-                        facecolor=color, alpha=base_alpha / i))
+        e.append(
+            ax.fill_between(
+                x,
+                mean - i * std,
+                mean + i * std,
+                facecolor=color,
+                alpha=base_alpha / i
+            )
+        )
     return (l, e)
+
+def summarize_sampler(sampler, burn=0, thin=1, ci=0.95):
+    r"""Create summary statistics of the flattened chain of the sampler.
+    
+    The confidence regions are computed from the quantiles of the data.
+    
+    Parameters
+    ----------
+    sampler : :py:class:`emcee.EnsembleSampler` instance
+        The sampler to summarize the chains of.
+    burn : int, optional
+        The number of samples to burn from the beginning of the chain. Default
+        is 0 (no burn).
+    thin : int, optional
+        The step size to thin with. Default is 1 (no thinning).
+    ci : float, optional
+        A number between 0 and 1 indicating the confidence region to compute.
+        Default is 0.95 (return upper and lower bounds of the 95% confidence
+        interval).
+    
+    Returns
+    -------
+    mean : array, (num_params,)
+        Mean values of each of the parameters sampled.
+    ci_l : array, (num_params,)
+        Lower bounds of the `ci*100%` confidence intervals.
+    ci_u : array, (num_params,)
+        Upper bounds of the `ci*100%` confidence intervals.
+    """
+    flat_trace = sampler.chain[:, burn::thin, :]
+    flat_trace = flat_trace.reshape((-1, flat_trace.shape[2]))
+    
+    mean = scipy.mean(flat_trace, axis=0)
+    cibdry = 100.0 * (1.0 - ci) / 2.0
+    ci_l, ci_u = scipy.percentile(flat_trace, [cibdry, 100.0 - cibdry], axis=0)
+    
+    return (mean, ci_l, ci_u)
+
+def plot_sampler(sampler, labels=None, burn=0):
+    """Plot the results of MCMC sampler (posterior and chains).
+    """
+    
+    # Loosely based on triangle.py
+    
+    # Create axes:
+    k = sampler.flatchain.shape[1]
+    
+    if labels is None:
+        labels = [''] * k
+    
+    f = plt.figure()
+    gs1 = mplgs.GridSpec(k, k)
+    gs2 = mplgs.GridSpec(1, k)
+    gs1.update(bottom=0.275, top=0.98)
+    gs2.update(bottom=0.1, top=0.2)
+    axes = []
+    # j is the row, i is the column.
+    for j in xrange(0, k + 1):
+        row = []
+        for i in xrange(0, k):
+            if i > j:
+                row.append(None)
+            else:
+                sharey = row[-1] if i > 0 and i < j and j < k else None
+                sharex = axes[-1][i] if j > i and j < k else \
+                    (row[-1] if i > 0 and j == k else None)
+                gs = gs1[j, i] if j < k else gs2[:, i]
+                row.append(f.add_subplot(gs, sharey=sharey, sharex=sharex))
+        axes.append(row)
+    axes = scipy.asarray(axes)
+    
+    # Update axes with the data:
+    flat_trace = sampler.chain[:, burn:, :]
+    flat_trace = flat_trace.reshape((-1, k))
+    
+    # j is the row, i is the column.
+    for i in xrange(0, k):
+        axes[i, i].clear()
+        axes[i, i].hist(flat_trace[:, i], bins=50, color='black')
+        if i == k - 1:
+            axes[i, i].set_xlabel(labels[i])
+        if i < k - 1:
+            plt.setp(axes[i, i].get_xticklabels(), visible=False)
+        plt.setp(axes[i, i].get_yticklabels(), visible=False)
+        # for j in xrange(0, i):
+        #     axes[j, i].set_visible(False)
+        #     axes[j, i].set_frame_on(False)
+        for j in xrange(i + 1, k):
+            axes[j, i].clear()
+            ct, x, y, im = axes[j, i].hist2d(flat_trace[:, i], flat_trace[:, j], bins=50, cmap='gray_r')
+            # xmid = 0.5 * (x[1:] + x[:-1])
+            # ymid = 0.5 * (y[1:] + y[:-1])
+            # axes[j, i].contour(xmid, ymid, ct.T, colors='k')
+            if j < k - 1:
+                plt.setp(axes[j, i].get_xticklabels(), visible=False)
+            if i != 0:
+                plt.setp(axes[j, i].get_yticklabels(), visible=False)
+            if i == 0:
+                axes[j, i].set_ylabel(labels[j])
+            if j == k - 1:
+                axes[j, i].set_xlabel(labels[i])
+        axes[-1, i].clear()
+        axes[-1, i].plot(sampler.chain[:, :, i].T)
+        axes[-1, i].axvline(burn, color='r', linewidth=3)
+        axes[-1, i].set_ylabel(labels[i])
+        axes[-1, i].set_xlabel('step')
+    
+    f.canvas.draw()
